import click
import urllib3
import threading
import os
from typing import Optional

from covid_cloud.constants import *
from requests.exceptions import HTTPError
from covid_cloud.client.utils import get_drs_client


def is_drs_url(url):
    return url[:3] == 'drs'


def download_file(drs_url, url, output_dir, oauth_token: Optional[str] = None):
    drs_client = get_drs_client(drs_url, oauth_token)

    http = urllib3.PoolManager()
    chunk_size = 1024
    download_url = url
    signed_access_ids = ['az-blobstore-signed']

    object_id = url.split('/')[-1]
    if is_drs_url(url):
        # parse the drs url to the resource url
        try:
            object_info = drs_client.get_object_info(object_id)
        except HTTPError as e:
            click.echo(e)
            if '404' in str(e.response):
                click.secho("There was an error getting object into from the DRS Client", fg='red')
            return
        except:
            return

        if "access_methods" in object_info.keys():
            for access_method in object_info["access_methods"][0]:
                if access_method.get('access_id', None):
                    if access_method['access_id'] in signed_access_ids:
                        click.echo("found signed access_id @ access_method level")
                        try:
                            object_access = drs_client.get_object_access(object_id, access_method['access_id'])
                        except HTTPError as e:
                            click.echo(e)
                            return
                        download_url = object_access['url']
                        break

                # if we have an https, use that
                if access_method['type'] == 'https':
                    download_url = access_method['access_url']['url']
                    break
        else:
            return  # next page token, just return

    try:
        download_stream = http.request('GET', download_url, preload_content=False)
    except:
        click.secho("There was an error downloading " + download_url, fg='red')
        return

    with open(f"{output_dir}/{download_url.split('/')[-1]}", 'wb+') as dest:
        while True:
            data = download_stream.read(chunk_size)
            if not data:
                break
            dest.write(data)


def download_files(drs_url, urls, output_dir=downloads_directory, oauth_token: Optional[str] = None):
    download_threads = []

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    for url in urls:
        download = threading.Thread(target=download_file(drs_url, url, output_dir, oauth_token), name=url)
        download.daemon = True
        download_threads.append(download)
        download.start()

    for thread in download_threads:
        thread.join()

    click.secho("Download Complete into " + output_dir, fg='green')



import os

wallet_uri = 'https://wallet.staging.dnastack.com' #TODO: replace with prod
redirect_uri = "https://92a15358-9be7-4edf-8758-ec87ee1f81e3.mock.pstmn.io/" #TODO: replace this with a proper DNAstack URI

config_file_path = f"{os.path.expanduser('~')}/.covid-cloud/config.yaml"
cli_directory = f"{os.path.expanduser('~')}/.covid-cloud"
downloads_directory = f"{os.path.expanduser('~')}/.covid-cloud/downloads"

ACCEPTED_CONFIG_KEYS = ['search-url', 'drs-url', 'personal_access_token', 'email', 'oauth_token', 'oauth_token_expiry',
                        'wallet-url', 'client-id', 'client-secret', 'client-redirect-uri']



import requests as req
from urllib.parse import parse_qs
from covid_cloud.constants import *


def login(email, personal_access_token, client_id, client_secret, client_redirect_uri, wallet_uri=wallet_uri,
          search_url='', drs_url=''):
    session = req.Session()

    # login at /login/token
    session.get(f'{wallet_uri}/login/token',
                params={
                    'token': personal_access_token,
                    "email": email,
                },
                allow_redirects=False
                )

    auth_code_res = session.get(f'{wallet_uri}/oauth/authorize',
                                params={
                                    'response_type': 'code',
                                    'client_id': client_id,
                                    'redirect_uri': client_redirect_uri,
                                    'scopes': 'openid,drs-object:write,drs-object:access',
                                    'resource': f'{drs_url},{search_url}'
                                },
                                allow_redirects=False)

    auth_code = parse_qs(req.utils.urlparse(auth_code_res.headers['Location']).query)['code'][0]

    auth_token_res = session.post(f'{wallet_uri}/oauth/token',
                                  data={
                                      'grant_type': 'authorization_code',
                                      'code': auth_code,
                                      'client_id': client_id,
                                      'client_secret': client_secret
                                  }
                                  )

    json_res = auth_token_res.json()

    return json_res



import click
import datetime as dt
from covid_cloud.cli.utils import has_config, get_config
from covid_cloud.cli.auth.commands import cli_login

def assert_token(ctx):

    # there is no good way got telling if the server is authenticated yet,
    # so if the user does not have a token set up, assume that is is public and
    # don't do anything
    if not has_config(ctx,"oauth_token"):
        return

    oauth_token = get_config(ctx,"oauth_token")
    if oauth_token["expiry"] > dt.datetime.now(tz=dt.timezone.utc):
        cli_login(ctx, refresh=True)


from covid_cloud.client import *
from covid_cloud import constants


class CovidCloudClient:
    def __init__(self, client_id, client_secret, personal_access_token, search_url=None, drs_url=None):
        self.client_id = client_id
        self.client_secret = client_secret
        self.personal_access_token = personal_access_token
        self.search_url = search_url
        self.drs_url = drs_url
        self.oauth_token = None

    def login(self, email, personal_access_token=None, client_id=None, client_secret=None, client_redirect_uri=None,
              wallet_uri=None, search_url=None, drs_url=None):
        return login(email=email, personal_access_token=self.personal_access_token, client_id=self.client_id,
                     client_secret=self.client_secret, client_redirect_uri=constants.redirect_uri,
                     wallet_uri=constants.wallet_uri, search_url=self.search_url, drs_url=self.drs_url)

    def query(self, q, download=False, use_json=False, raw=False):
        return query(self.search_url, q, download, use_json, raw, self.oauth_token)

    def list_tables(self):
        return get_tables(self.search_url, self.oauth_token)

    def get_table(self, table_name):
        return get_table(self.search_url, table_name, self.oauth_token)

    def download(self, urls, output_dir=downloads_directory):
        return download_files(self.drs_url, urls, output_dir, self.oauth_token)



import click
from covid_cloud.client.auth import login
from covid_cloud.client import *
from covid_cloud.cli.utils import assert_config, get_config, set_config
import covid_cloud.constants
import yaml
import datetime as dt


@click.group()
@click.pass_context
def auth(ctx):
    pass


@auth.command("login")
@click.pass_context
def cli_login(ctx, refresh=False):
    personal_access_token = get_config(ctx, "personal_access_token")
    email = get_config(ctx, "email")
    client_id = get_config(ctx, "client-id")
    client_secret = get_config(ctx, "client-secret")
    client_redirect_uri = get_config(ctx, "client-redirect-uri")

    try:
        access_token = login(email, personal_access_token, client_id, client_secret, client_redirect_uri,
                             wallet_uri=get_config(ctx, 'wallet-url'), search_url=get_config(ctx, 'search-url'),
                             drs_url=get_config(ctx, "drs-url"))
    except:
        click.secho("There was an error generating an access token", fg='red')
        return

    # convert the expiry from a relative time to an actual timestamp
    access_token['expiry'] = (
                dt.datetime.now(tz=dt.timezone.utc) + dt.timedelta(minutes=access_token['expires_in'])).timestamp()
    del access_token['expires_in']

    set_config(ctx, 'oauth_token', access_token)

    # if we're refrshing the token, we don't want the user to see that we're refreshing
    if not refresh:
        click.secho('You are now logged in.', fg='green')
